from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.programming.output import Output
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.downloader import Downloader
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.env import Environment
from typing import Union

import time
import requests


PRODIA_API_KEY =  Environment.get_current_project_env('PRODIA_API_KEY')

class Prodia:
    """
    Class to wrap prodia functionality. Prodia is an
    image generator engine.
    """

    @staticmethod
    def generate_image(
        prompt: str,
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        if not PythonValidator.is_string(prompt):
            raise Exception('The provided "prompt" is not a valid string.')
        
        # If you comment this and uncomment the one below it works
        # seed = randint(1000000000, 9999999999)
        # response = requests.get('https://api.prodia.com/generate?new=true&prompt=' + prompt + '&model=absolutereality_v181.safetensors+%5B3d9d4d2b%5D&steps=20&cfg=7&seed=' + str(seed) + '&sampler=DPM%2B%2B+2M+Karras&aspect_ratio=square')
        payload = {
            'new': True,
            'prompt': prompt,
            #'model': 'absolutereality_v181.safetensors [3d9d4d2b]',   # this model works on above request, not here
            'model': 'sd_xl_base_1.0.safetensors [be9edd61]',
            #'negative_prompt': '',
            'steps': 20,
            'cfg_scale': 7,
            'seed': 2328045384,
            'sampler': 'DPM++ 2M Karras',
            'width': 1344,
            'height': 768
        }
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "X-Prodia-Key": PRODIA_API_KEY
        }
        url = 'https://api.prodia.com/v1/sdxl/generate'
        response = requests.post(url, json = payload, headers = headers)
        response = response.json()

        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)

        # When requested it is queued, so we ask for it until it is done
        if "status" in response and response['status'] == 'queued':
            job_id = response['job']
            return Prodia._retrieve_job(job_id, output_filename)
        else:
            print(response)
            raise Exception('There was an error when generating a Prodia AI Image.')

    @staticmethod
    def _retrieve_job(
        job_id: str,
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        Makes a request for the image that is being
        generated with the provided 'job_id'.

        It has a loop to wait until it is done. This
        code is critic because of the loop.
        """
        url = f'https://api.prodia.com/v1/job/{str(job_id)}'

        headers = {
            'accept': 'application/json',
            'X-Prodia-Key': PRODIA_API_KEY
        }

        response = requests.get(url, headers = headers)
        response = response.json()
        #print(response)

        # TODO: Do a passive waiting
        is_downloadable = True

        if response['status'] != 'succeeded':
            is_downloadable = False

        # TODO: Implement a tries number
        while not is_downloadable:
            time.sleep(5)
            print('Doing a request in loop')

            # We do the call again
            response = requests.get(url, headers = headers)
            response = response.json()
            print(response)
            if 'imageUrl' in response:
                is_downloadable = True

        image_url = response['imageUrl']

        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)

        return Downloader.download_image(image_url, output_filename)


from yta_general_utils.text_processor import remove_marks_and_accents
from utils.ai.chat import get_script

CHARS_PER_SECOND = 17 # For Matthew
SECONDS_PER_PAUSE = 0.4 # Estimated by me

# TODO: Here I will generate narration scripts from an idea, from cero
def __seconds_to_chars(seconds):
    """
    Returns the number of chars that, when narrated, will last the expected 'seconds'
    duration in seconds.
    """
    # This has been manually calculated for Matthew. Each voice will work at its own speed.
    # 1 second => 17 characters
    # 1 '.' => +0.5 seconds
    return int(CHARS_PER_SECOND * seconds)

def __chars_to_seconds(text):
    """
    Returns the number of seconds that the provided chars number would last in a 
    narration.
    """
    return len(text) / CHARS_PER_SECOND

def __estimate_duration(text: str):
    """
    This method analyzes the provided narration script to estimate the duration that
    the text will last when narrated.
    """
    # We estimate based on simple chars without marks. Only narrative chars
    estimated_seconds = __chars_to_seconds(remove_marks_and_accents(text))
    # We treat each stop mark as a pause, so we add that time in seconds
    count = 0
    count += text.count('?')
    text = text.replace('?', '')
    count += text.count('!')
    text = text.replace('!', '')
    count += text.count('...')
    text = text.replace('...', '')
    count += text.count('.')
    text = text.replace('.', '')
    estimated_seconds += SECONDS_PER_PAUSE * count

    return estimated_seconds


def generate(idea, duration):
    """
    Generates a narration script from the provided 'idea' that will last around 'duration' minutes.

    # TODO: Finish it
    """
    # TODO: Ask AI for an script of 'x' characters

    # I want to create one script that lasts this time below
    time_in_seconds = 60 * duration
    needed_chars = __seconds_to_chars(time_in_seconds)

    # TODO: Ask for AI generated script
    return 'This need to be done, by now is just a fixed text.'
    return get_script(idea, needed_chars)



# https://www.reddit.com/r/OpenAI/comments/vm3kfk/comment/ie4j7vf/


"""
Flux is an image generation software created by Blackforest Labs
(https://blackforestlabs.ai/). They have 3 different models and
2 of those are free-to-use.

Those models ('dev' and 'schnell') have been implemented here to
be used to generate images.

The HugginFace endpoints I've found when creating this file:
- https://huggingface.co/black-forest-labs/FLUX.1-schnell?inference_api=true
- https://huggingface.co/black-forest-labs/FLUX.1-dev?inference_api=true
"""
from yta_ai_utils.hugginface import HugginFace
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union


SCHNELL_HUGGINFACE_ENDPOINT = 'https://api-inference.huggingface.co/models/black-forest-labs/FLUX.1-schnell'
DEV_HUGGINFACE_ENDPOINT = 'https://api-inference.huggingface.co/models/black-forest-labs/FLUX.1-dev'

class Flux:
    """
    Class to wrap the image generation with 
    Flux engine.
    """

    @staticmethod
    def generate_dev(
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        """
        Generates an image using the Flux 'dev' model, through the 
        HugginFace endpoint, and returns it read as PIL image. If
        'output_filename' provided, it will be also stored locally
        with that name.

        @param
            **prompt**
            The prompt you want to use in the image generation. The
            more specific and explicit you are, the more accurate 
            result you get.

        @param
            **output_filename**
            The filename you want for the image to be stored locally.
            If None provided it won't be stored.
        """
        if not PythonValidator.is_string(prompt):
            raise Exception('No valid "prompt" provided.')
        
        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
        
        return HugginFace(DEV_HUGGINFACE_ENDPOINT).generate_image({'inputs': prompt}, output_filename)

    @staticmethod
    def generate_schnell(
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        """
        Generates an image using the Flux 'dev' model, through the 
        HugginFace endpoint, and returns it read as PIL image. If
        'output_filename' provided, it will be also stored locally
        with that name.

        @param
            **prompt**
            The prompt you want to use in the image generation. The
            more specific and explicit you are, the more accurate 
            result you get.

        @param
            **output_filename**
            The filename you want for the image to be stored locally.
            If None provided it won't be stored.
        """
        if not PythonValidator.is_string(prompt):
            raise Exception('No valid "prompt" provided.')
        
        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
        
        return HugginFace(SCHNELL_HUGGINFACE_ENDPOINT).generate_image({'inputs': prompt}, output_filename)

# TODO: Try to download the models and implement with Diffusers
# https://huggingface.co/black-forest-labs/FLUX.1-schnell?inference_api=true
# https://huggingface.co/black-forest-labs/FLUX.1-schnell?inference_api=true

# Interesting: https://www.youtube.com/watch?v=P-rzgaIfZCo


from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.downloader import Downloader
from typing import Union

from openai import OpenAI


# TODO: Is this actually useful? I think it could be removed...
class OpenAI:
    """
    Class to wrap the OpenAI functionality.
    """
    
    def generate_image(
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        if not PythonValidator.is_string(prompt):
            raise Exception('The provided "prompt" is not a valid string.')
        
        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)

        client = OpenAI()

        response = client.images.generate(
            model = "dall-e-3",
            prompt = prompt,
            size = "1792x1024",
            quality = "standard",
            n = 1,
        )

        image_url = response.data[0].url

        return Downloader.download_image(image_url, output_filename)


from yta_general_utils.programming.env import Environment
from yta_general_utils.programming.parameter_validator import PythonValidator

import google.generativeai as genai


GEMINI_API_KEY =  Environment.get_current_project_env('GEMINI_API_KEY')

class Gemini:
    """
    Class to wrap the Gemini AI chatbot.
    """

    @staticmethod
    def ask(
        prompt: str
    ):
        """
        Ask Gemini AI (gemini-1.5-flash) model by using the
        provided 'prompt' and return its response.
        """
        if not PythonValidator.is_string(prompt):
            raise Exception('The provided "prompt" is not a valid string.')

        genai.configure(api_key = GEMINI_API_KEY)
        model = genai.GenerativeModel('gemini-1.5-flash')

        chat = model.start_chat()
        response = chat.send_message(
            prompt,
        )

        return response.text


# I found this (https://github.com/f/awesome-chatgpt-prompts) that could help

def get_image_generation_input_prompt(idea):
    """
    Returns the input prompt to ask GPT for a prompt that would use in midjourney to generate
    an incredible AI image based  on the provided 'idea'.
    
    Explicitly requests that only the prompt be responded to, without any additions.
    """
    return "¿Cuál sería el prompt que utilizarías para que MidJourney te crease una imagen que represente la idea '" + idea + "'? Respóndeme única y exclusivamente con el prompt, sé conciso."

def get_stock_video_input_prompt(idea):
    """
    Returns the input prompt to ask GPT for a key phrase that would use in YouTube to search
    for the perfect video to fit the provided 'idea'.

    Explicitly requests that only the key phrase be responded to, without any additions.
    """
    #return "quiero que me seas conciso y no me digas nada más que no sea qué palabras clave utilizarías para buscar un vídeo en Youtube que pueda servir para poner de fondo en un vídeo y que represente la frase '" + idea + "'"

    return "¿Cuál sería una frase que podrías usar para buscar en YouTube un vídeo de stock que pueda poner de fondo en otro vídeo mientras hablo de " + idea + " y que en dicho vídeo se represente eso de lo que estoy hablando, como si fuera un vídeo de stock? Respóndeme única y exclusivamente con la frase, sé conciso."

def get_feeling_from_input_prompt(input):
    """
    Returns the input prompt to ask GPT for the feeling that the provided 'input' (that 
    is a text) that conveys.

    Explicitly requests that only the feeling be responded to, without any additions.
    """
    return "Me han dicho la frase " + input + " y necesito saber qué tipo de sentimiento transmite esa frase. Solo quiero que me respondas el sentimiento, única y exclusivamente, así que sé concisco. No me des más de una opción. Dame solo una opción."

def get_topic_from_input_prompt(input):
    """
    Returns the input prompt to ask GPT for the main topic that is being talked about in
    the provided 'input'.

    Explicitly requests that only the topic be responded to, without any additions.
    """
    return "¿Me puedes cuál es el tema principal del que se habla en la frase '" + input + "'? Solamente respóndeme con el tema, sé conciso. No me des más de una opción. Dame solo una opción."

def get_translate_text_input_prompt(input, input_language = 'inglés', output_language = 'español'):
    """
    Returns the input prompt to ask GPT for a translation from 'input_language' to 'output_language'
    of the provided 'input'.

    Explictly requests that only the translation be responded to, without any additions.
    """
    return "¿Puedes traducirme el texto que te voy a pasar a continuación del " + input_language + " al " + output_language + "? Solo quiero que me respondas con el texto traducido, sé conciso. No añadas nada más que el texto. El texto es el siguiente, que te lo mando entrecomillado. '" + input + "'"

def get_script_prompt(input, num_of_chars):
    return 'Quiero que actúes como si fueras un escritor profesional muy bueno de guiones de vídeos de Youtube. Sabes generar guiones que mantienen la atención del espectador durante todo el tiempo, y el guion es tan espectacular que lo querrían las principales empresas de Hollywood de lo bueno que es. Te voy a dar la idea sobre la que tratará el guion y tú me respondes única y exclusivamente con ese guión. El guion debe contar con un total de ' + str(num_of_chars) + ' caracteres, y la idea es "' + input + '".'

def get_youtube_summary_from_subtitles_prompt(input):
    #return 'Tengo los subtítulos de un vídeo de Youtube. ¿Me puedes hacer un resumen de lo que pasa en el vídeo? Quiero que el resumen sea por capítulos. No quiero que me digas nada más que el resumen en sí, pero que este sea bastante detallado en cada capítulo. Estos son los subtítulos: "' + input + '"'

    #return 'Te voy a dar los subtítulos de un vídeo de Youtube. Quiero que me hagas un resumen de unos mil caracteres de largo, todo en un mismo párrafo, sin saltos de línea, como si fuese una historia. Estos son los subtítulos: "' + input + '"'

    return 'Tengo los subtítulos de un vídeo. Quiero que me expliques lo que pasa en el vídeo, sin saltos de línea ni títulos ni nada, en un solo párrafo, como si fuese una historia. No seas escueto. Estos son los subtítulos: "' + input + '"'

    # This one returns me line breaks, asterisks, titles... It works for a human, but not for me
    #return 'Quiero que actúes como si fueras alguien que se dedica a crear vídeos en Youtube. Eres experto en crear guiones, a modo de historia, para vídeos de Youtube. Te voy a dar los subtítulos de un vídeo y quiero que me reconstruyas un guion que sirva para narrar otro vídeo nuevo sobre lo que se habla en el vídeo del que te paso los subtítulos a continuación. Solamente quiero que me des como respuesta el guion, sin títulos, sin líneas de separación. Todo seguido en el mismo párrafo, aunque sea muy muy largo. No quiero saltos de línea ni títulos, todo seguido, en la misma línea. Los subtítulos son los siguientes: "' + input + '"'

def get_story(input):
    """
    Returns a story from a professional storyteller that talks about the provided 'input' idea.
    It comes in different paragraphs, so it could be a good idea to use each one as a segment.
    """
    return 'Quiero que actúes como si fueras un narrador de historias, un storyteller. Vas a ser capaz de inventar increíbles historias que tienen gancho, que son imaginativas y que mantienen la atención de la audiencia hasta el final. Pueden ser historias educacionales, inventadas, basadas en hechos reales o cualquier tipo de historias que tengan el potencial de capturar la atención e imaginación de la gente. Dependiendo del tipo de audiencia, vas a elegir diferentes temas para tu historia. Por ejemplo, si son niños puedes hablar sobre animales. Si son adultos, puedes hablar sobre historias basadas en hechos reales que cautiven mejor a esa audiencia. Mi primera petición es: "' + input + '"'

def get_english_song(topic, chars_length):
    return 'Quiero que actúes como si fueses un increíble compositor de canciones en inglés y que me des la letra de una canción. Quiero que la canción tenga máximo ' + str(chars_length) + ' caracteres, que trate sobre ' + topic + ', y que esté en inglés. Respóndeme solo con la letra de la canción directamente.'


from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.downloader import Downloader
from yta_general_utils.programming.env import Environment
from typing import Union

import requests
import os
import time


MODEL_LEONARDO_DIFFUSION_XL = '1e60896f-3c26-4296-8ecc-53e2afecc132'

class Leonardo:
    """
    Class to wrap the functionality of Leonardo
    AI.
    """

    @staticmethod
    def _request_generation(
        prompt: str
    ):
        """
        Makes a request to generate an image. It returns the generation id.

        This method needs the non-free API working.
        """
        if not PythonValidator.is_string(prompt):
            raise Exception('The provided "prompt" is not a valid string.')
        
        #return '69503ef1-85c8-41cc-b292-f672946920d7'
        url = "https://cloud.leonardo.ai/api/rest/v1/generations"

        payload = {
            'modelId': MODEL_LEONARDO_DIFFUSION_XL,
            'width': 1024,   # test 1536
            'height': 576,   # test 864
            'num_images': 1,
            'prompt': prompt
        }

        LEONARDO_API_KEY = Environment.get_current_project_env('LEONARDOAI_API_KEY')

        headers = {
            "accept": "application/json",
            "authorization": f'Bearer {LEONARDO_API_KEY}',
            "content-type": "application/json",
        }

        response = requests.post(url, json = payload, headers = headers)
        response = response.json()

        print(response)

        return response['sdGenerationJob']['generationId']

    @staticmethod
    def _download_generated_image(generation_id, output_filename):
        """
        Downloads the AI-generated image by using the provided 'generation_id'.

        It downloads the image and resizes it to 1920x1080.
        """
        url = 'https://cloud.leonardo.ai/api/rest/v1/generations/' + str(generation_id)

        headers = {
            "accept": "application/json",
            "authorization": "Bearer " + str(os.getenv('LEONARDOAI_API_KEY')),
        }

        response = requests.get(url, headers = headers)
        response = response.json()

        # TODO: Do a passive waiting
        is_downloadable = True

        if len(response['generations_by_pk']['generated_images']) == 0:
            is_downloadable = False

        while not is_downloadable:
            time.sleep(10)
            print('Doing a request in loop')

            # We do the call again
            response = requests.get(url, headers = headers)
            response = response.json()
            
            if len(response['generations_by_pk']['generated_images']) > 0:
                is_downloadable = True

        downloadable_url = response['generations_by_pk']['generated_images'][0]['url']

        return Downloader.download_image(downloadable_url, output_filename)

    @staticmethod
    def generate_image(
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        """
        Generate an AI image with the provided 'prompt' and
        store it locally as the given 'output_filename'.
        """
        if not PythonValidator.is_string(prompt):
            raise Exception('The provided "prompt" is not a valid string.')
        
        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
        
        return Leonardo._download_generated_image(
            Leonardo._request_generation(prompt),
            output_filename
        )

